/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1996, 1997
 *	Sleepycat Software.  All rights reserved.
 */
/*
 * Copyright (c) 1990, 1993, 1994, 1995, 1996
 *	Keith Bostic.  All rights reserved.
 */
/*
 * Copyright (c) 1990, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Olson.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)bt_close.c	10.18 (Sleepycat) 7/26/97";
#endif /* not lint */

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>
#include <sys/mman.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif

#include "db_int.h"
#include "db_page.h"
#include "btree.h"

static int __ram_writeback __P((DB *));

/*
 * __bam_close --
 *	Close a btree.
 *
 * PUBLIC: int __bam_close __P((DB *));
 */
int
__bam_close(dbp)
	DB *dbp;
{
	BTMETA *meta;
	BTREE *t;
	DB_LOCK mlock;
	db_pgno_t pgno;
	int flags, ret, t_ret;

	DEBUG_LWRITE(dbp, NULL, "bam_close", NULL, NULL, 0);

	t = dbp->internal;
	ret = 0;

	/*
	 * Update the tree statistics.
	 *
	 * XXX
	 * We don't log the change for updating the statistics on the metadata
	 * page.  It's small enough that corrupted data is probably impossible,
	 * but we may lose information.
	 */
	if (!F_ISSET(dbp, DB_AM_INMEM | DB_AM_RDONLY)) {
		flags = 0;
		pgno = PGNO_METADATA;
		if (__bam_lget(dbp, 0, pgno, DB_LOCK_WRITE, &mlock) == 0) {
			if (__bam_pget(dbp, (PAGE **)&meta, &pgno, 0) == 0) {
				flags = DB_MPOOL_DIRTY;
				__bam_add_gstat(&t->lstat, &meta->stat);
				(void)memp_fput(dbp->mpf, (PAGE *)meta, flags);
			}
			(void)__BT_LPUT(dbp, mlock);
		}
	}

	/*
	 * Close underlying mmap region and backing source file descriptor,
	 * and free the name.
	 *
	 * XXX
	 * Safe -- DB_THREAD can't be set if there's a backing recno file.
	 */
	if (!F_ISSET(dbp, DB_AM_THREAD)) {
		if (t->re_smap != NULL) {
			(void)__db_munmap(t->re_smap, t->re_msize);
			t->re_smap = NULL;
		}
		if (t->re_fd != -1 &&
		    (t_ret = __db_close(t->re_fd)) != 0 && ret == 0)
			ret = t_ret;
		if (t->re_source != NULL)
			FREES(t->re_source);
	}

	/* Free any allocated memory. */
	if (t->bt_rkey.data) {
		FREE(t->bt_rkey.data, t->bt_rkey.size);
		t->bt_rkey.size = 0;
		t->bt_rkey.data = NULL;
	}
	if (t->bt_rdata.data) {
		FREE(t->bt_rdata.data, t->bt_rdata.ulen);
		t->bt_rdata.ulen = 0;
		t->bt_rdata.data = NULL;
	}

	if (t->bt_sp != t->bt_stack)
		FREE(t->bt_sp, (t->bt_esp - t->bt_sp) * sizeof(EPG));

	FREE(t, sizeof(BTREE));
	dbp->internal = NULL;

	return (ret);
}

/*
 * __bam_sync --
 *	Sync the btree to disk.
 *
 * PUBLIC: int __bam_sync __P((DB *, int));
 */
int
__bam_sync(argdbp, flags)
	DB *argdbp;
	int flags;
{
	BTREE *t;
	DB *dbp;
	int ret;

	DEBUG_LWRITE(argdbp, NULL, "bam_sync", NULL, NULL, flags);

	/* Check for invalid flags. */
	if ((ret = __db_syncchk(argdbp, flags)) != 0)
		return (ret);

	GETHANDLE(argdbp, NULL, &dbp, ret);
	t = dbp->internal;

	/*
	 * XXX
	 * Safe -- DB_THREAD can't be set if there's a backing recno file.
	 */
	if (dbp->type == DB_RECNO && (ret = __ram_writeback(dbp)))
		goto err;

	/* If it wasn't possible to modify the file, we're done. */
	if (F_ISSET(dbp, DB_AM_INMEM | DB_AM_RDONLY)) {
		PUTHANDLE(dbp);
		return (0);
	}

	/* Flush any dirty pages from the cache to the backing file. */
	ret = memp_fsync(dbp->mpf);

err:	PUTHANDLE(dbp);
	return (ret);
}

/*
 * __ram_writeback --
 *	Rewrite the backing file.
 */
static int
__ram_writeback(dbp)
	DB *dbp;
{
	BTREE *t;
	DBC *dbcp;
	DBT key, data;
	db_recno_t keyno;
	ssize_t nw;
	int fd, ret;
	u_int8_t delim;

	t = dbp->internal;

	/*
	 * If no modifications were made or there's no backing source file,
	 * we're done.
	 */
	if (!F_ISSET(dbp, DB_RE_MODIFIED) || t->re_fd == -1)
		return (0);
	F_CLR(dbp, DB_RE_MODIFIED);

	/*
	 * Read any remaining records into the tree.
	 *
	 * XXX
	 * This is why we can't support transactions when applications specify
	 * backing (re_source) files.  At this point we have to read in the
	 * rest of the records from the file so that we can write all of the
	 * records back out again, which could modify a page for which we'd
	 * have to log changes and which we don't have locked.  This could be
	 * partially fixed by taking a snapshot of the entire file during the
	 * db_open(), or, since db_open() isn't transaction protected, as part
	 * of the first DB operation.  But, if a checkpoint occurs then, the
	 * part of the log holding the copy of the file could be discarded, and
	 * that would make it impossible to recover in the face of disaster.
	 * This could all probably be fixed, but it would require transaction
	 * protecting the backing source file, i.e. mpool would have to know
	 * about it.
	 */
	if ((ret =
	    __ram_update(dbp, DB_MAX_RECORDS, 0)) != 0 && ret != DB_NOTFOUND)
		return (ret);

	/*
	 * !!!
	 * Close any underlying mmap region.  This is required for Windows NT
	 * (4.0, Service Pack 2) -- if the file is still mapped, the following
	 * open will fail.
	 */
	if (t->re_smap != NULL) {
		(void)__db_munmap(t->re_smap, t->re_msize);
		t->re_smap = NULL;
	}

	/* Open the file, truncating it. */
	if ((ret = __db_fdopen(t->re_source,
	    DB_SEQUENTIAL | DB_TRUNCATE,
	    DB_SEQUENTIAL | DB_TRUNCATE, 0, &fd)) != 0) {
		__db_err(dbp->dbenv, "%s: %s", t->re_source, strerror(ret));
		return (ret);
	}

	/* Get a cursor and write the records out. */
	if ((ret = dbp->cursor(dbp, NULL, &dbcp)) != 0)
		goto err;

	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	key.size = sizeof(db_recno_t);
	key.data = &keyno;
	if (F_ISSET(dbp, DB_BT_FIXEDLEN)) {
		while (!(ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT))) {
			if ((ret = __db_write(t->re_fd,
			    data.data, data.size, &nw)) != 0)
				goto err;
			if (nw != (ssize_t)data.size)
				goto eio;
		}
	} else {
		delim = t->re_delim;
		while (!(ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT))) {
			if ((ret = __db_write(t->re_fd,
			    data.data, data.size, &nw)) != 0)
				goto err;
			if (nw != (ssize_t)data.size)
				goto eio;
			if ((ret =
			    __db_write(t->re_fd, &delim, 1, &nw)) != 0)
				goto err;
			if (nw != 1) {
eio:				ret = EIO;
				goto err;
			}
		}
	}
	/* Close the file descriptor. */
	if ((ret = __db_close(fd)) != 0 || ret != 1)
		goto err;

	/* Discard the cursor. */
	(void)dbcp->c_close(dbcp);

	return (0);

err:	(void)__db_close(fd);
	(void)dbcp->c_close(dbcp);
	return (ret);
}
