/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1996, 1997
 *	Sleepycat Software.  All rights reserved.
 */
/*
 * Copyright (c) 1990, 1993, 1994, 1995, 1996
 *	Keith Bostic.  All rights reserved.
 */
/*
 * Copyright (c) 1990, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Olson.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)bt_get.c	10.9 (Sleepycat) 7/26/97";
#endif /* not lint */

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "db_int.h"
#include "db_page.h"
#include "btree.h"

/*
 * __ram_get --
 *	Retrieve a key/data pair from the tree.
 *
 * PUBLIC: int __ram_get __P((DB *, DB_TXN *, DBT *, DBT *, int));
 */
int
__ram_get(argdbp, txn, key, data, flags)
	DB *argdbp;
	DB_TXN *txn;
	DBT *key, *data;
	int flags;
{
	BTREE *t;
	DB *dbp;
	db_recno_t recno;
	int exact, ret;

	DEBUG_LWRITE(argdbp, txn, "ram_get", key, NULL, flags);

	/* Check for invalid flags. */
	if ((ret = __db_getchk(argdbp, key, data, flags)) != 0)
		return (ret);
	if ((ret = __ram_getno(argdbp, key, &recno)) != 0)
		return (ret);

	GETHANDLE(argdbp, txn, &dbp, ret);
	t = dbp->internal;

	/* Search the tree for the record. */
	if (!F_ISSET(dbp, DB_BT_EOF) && (ret = __ram_update(dbp, recno, 0)))
		return (ret);
	if ((ret = __ram_search(dbp, recno, S_FIND, 1, &exact)) != 0)
		return (ret);
	if (!exact)
		return (DB_NOTFOUND);

	/* Return the data item. */
	ret = __db_ret(dbp, t->bt_csp->page,
	    dbp->type == DB_RECNO ? t->bt_csp->indx : t->bt_csp->indx + 1,
	    data, &t->bt_rdata.data, &t->bt_rdata.ulen);
	++t->lstat.bt_get;

	/* Discard the page. */
	(void)memp_fput(dbp->mpf, t->bt_csp->page, 0);
	(void)__BT_TLPUT(dbp, t->bt_csp->lock);

	PUTHANDLE(dbp);
	return (ret);
}

/*
 * __ram_fmap --
 *	Get fixed length records from a file.
 *
 * PUBLIC: int __ram_fmap __P((DB *, db_recno_t));
 */
int
__ram_fmap(dbp, top)
	DB *dbp;
	db_recno_t top;
{
	BTREE *t;
	DBT data;
	db_recno_t recno;
	u_int32_t len;
	u_int8_t *sp, *ep, *p;
	int ret;

	if ((ret = __ram_nrecs(dbp, &recno)) != 0)
		return (ret);

	t = dbp->internal;
	if (t->bt_rdata.ulen < t->re_len) {
		t->bt_rdata.data = t->bt_rdata.data == NULL ?
		    (void *)malloc(t->re_len) :
		    (void *)realloc(t->bt_rdata.data, t->re_len);
		if (t->bt_rdata.data == NULL) {
			t->bt_rdata.ulen = 0;
			return (ENOMEM);
		}
		t->bt_rdata.ulen = t->re_len;
	}

	memset(&data, 0, sizeof(data));
	data.data = t->bt_rdata.data;
	data.size = t->re_len;

	sp = (u_int8_t *)t->re_cmap;
	ep = (u_int8_t *)t->re_emap;
	while (++recno <= top) {
		if (sp >= ep) {
			F_SET(dbp, DB_BT_EOF);
			return (DB_NOTFOUND);
		}
		len = t->re_len;
		for (p = t->bt_rdata.data;
		    sp < ep && len > 0; *p++ = *sp++, --len);
		if (len != 0)
			memset(p, t->re_pad, len);
		if ((ret = __ram_add(dbp, recno, &data, 0)) != 0)
			return (ret);
	}
	t->re_cmap = sp;
	return (0);
}

/*
 * __ram_vmap --
 *	Get variable length records from a file.
 *
 * PUBLIC: int __ram_vmap __P((DB *, db_recno_t));
 */
int
__ram_vmap(dbp, top)
	DB *dbp;
	db_recno_t top;
{
	BTREE *t;
	DBT data;
	db_recno_t recno;
	u_int8_t *sp, *ep;
	int delim, ret;

	t = dbp->internal;

	if ((ret = __ram_nrecs(dbp, &recno)) != 0)
		return (ret);

	memset(&data, 0, sizeof(data));

	delim = t->re_delim;

	sp = (u_int8_t *)t->re_cmap;
	ep = (u_int8_t *)t->re_emap;
	while (++recno <= top) {
		if (sp >= ep) {
			F_SET(dbp, DB_BT_EOF);
			return (DB_NOTFOUND);
		}
		for (data.data = sp; sp < ep && *sp != delim; ++sp);
		data.size = sp - (u_int8_t *)data.data;
		if ((ret = __ram_add(dbp, recno, &data, 0)) != 0)
			return (ret);
		++sp;
	}
	t->re_cmap = sp;
	return (0);
}

/*
 * __ram_getno --
 *	Get a record number from the user.
 *
 * PUBLIC: int __ram_getno __P((DB *, const DBT *, db_recno_t *));
 */
int
__ram_getno(dbp, key, rep)
	DB *dbp;
	const DBT *key;
	db_recno_t *rep;
{
	memcpy(rep, key->data, sizeof(db_recno_t));
	if (*rep == 0) {
		__db_err(dbp->dbenv, "illegal record number of 0");
		return (EINVAL);
	}
	return (0);
}

/*
 * __ram_total --
 *	Return the number of recno entries below a page.
 *
 * PUBLIC: db_recno_t __ram_total __P((PAGE *));
 */
db_recno_t
__ram_total(h)
	PAGE *h;
{
	db_recno_t recs;
	db_indx_t nxt, top;

	if (TYPE(h) == P_LRECNO)
		return (NUM_ENT(h));

	for (recs = 0, nxt = 0, top = NUM_ENT(h); nxt < top; ++nxt)
		recs += GET_RINTERNAL(h, nxt)->nrecs;
	return (recs);
}

/*
 * __ram_nrecs --
 *	Get the number of records in the tree.
 *
 * PUBLIC: int __ram_nrecs __P((DB *, db_recno_t *));
 */
int
__ram_nrecs(dbp, rep)
	DB *dbp;
	db_recno_t *rep;
{
	DB_LOCK lock;
	PAGE *h;
	db_pgno_t pgno;
	int ret;

	pgno = PGNO_ROOT;
	if ((ret = __bam_lget(dbp, 0, pgno, DB_LOCK_READ, &lock)) != 0)
		return (ret);
	if ((ret = __bam_pget(dbp, &h, &pgno, 0)) != 0)
		return (ret);

	*rep = RE_NREC(h);

	(void)memp_fput(dbp->mpf, h, 0);
	(void)__BT_TLPUT(dbp, lock);
	return (0);
}
