/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)cxx_app.cpp	10.7 (Sleepycat) 7/27/97";
#endif /* not lint */

#include "db_cxx.h"
#include "cxx_int.h"

#include <string.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>              // needed for setErrorStream

//
// Utility routine to allocate/duplicate a string
//
static char *dup(const char *s)
{
    char *result = 0;
    if (s) {
        result = new char[strlen(s) + 1];
        strcpy(result, s);
    }
    return result;
}

//
// Utility routine to allocate/duplicate a null terminated array of strings
//
static char **dup_strarray(char * const *sa)
{
    char **result = 0;
    if (sa) {
        int count = 0;
        char * const *pa;
        for (pa = sa; *pa; pa++)
            count++;
        typedef char *charstar;
        result = new charstar[count+1];
        for (int i=0; i<=count; i++)
            result[i] = dup(sa[i]);
    }
    return result;
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBVersion                               //
//                                                                    //
////////////////////////////////////////////////////////////////////////

static DBVersion version(DB_VERSION_MAJOR, DB_VERSION_MINOR, DB_VERSION_PATCH);

DBVersion::DBVersion(int major, int minor, int patch)
:   major_(major)
,   minor_(minor)
,   patch_(patch)
{
}

const char *DBVersion::GetVersionString()
{
    return DB_VERSION_STRING;
}

DBVersion &DBVersion::GetVersion()
{
    return version;
}

int DBVersion::getMajor() const
{
    return major_;
}

int DBVersion::getMinor() const
{
    return minor_;
}

int DBVersion::getPatch() const
{
    return patch_;
}


////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBApp                                   //
//                                                                    //
////////////////////////////////////////////////////////////////////////

const DBAppOptions DBApp::DefaultAppOptions;
const DBOpenBTreeParams DBApp::DefaultOpenBTreeParams;
const DBOpenHashParams DBApp::DefaultOpenHashParams;
const DBOpenRecnoParams DBApp::DefaultOpenRecnoParams;
const DBOpenUnknownParams DBApp::DefaultOpenUnknownParams;

static DBApp *currentApp = 0;

DBApp::DBApp(const char *homeDir, DBAppOptions options)
:   imp_(0)
,   errorModel_(DBAppOptions::Exception)
,   name_(0)
{
    // Make our own copy of the env in case the user destroys
    // theirs.
    //
    DB_ENV *option_env = unwrapOptions(&options);
    DB_ENV *env = new DB_ENV;
    *env = *option_env;
    env->db_errpfx = dup(env->db_errpfx);
    imp_ = (DBAppImp*)env;

    int flags = options.options_;
    errorModel_ = options.errorModel_;
    int err;

    if ((err = db_appinit(homeDir, 0, env, flags)) != 0) {
        DB_ERROR("DBApp constructor", err);
    }
    currentApp = this;
}

DBApp::DBApp(const char *homeDir, DBAppOptions options,
               const DBConfig &config)
:   imp_(0)
,   errorModel_(DBAppOptions::Exception)
,   name_(0)
{
    // Make our own copy of the env in case the user destroys
    // theirs.
    //
    DB_ENV *option_env = unwrapOptions(&options);
    DB_ENV *env = new DB_ENV;
    *env = *option_env;
    env->db_errpfx = dup(env->db_errpfx);
    imp_ = (DBAppImp*)env;

    int flags = options.options_;
    errorModel_ = options.errorModel_;
    int err;

    if ((err = db_appinit(homeDir, config.params_, env, flags)) != 0) {
        DB_ERROR("DBApp constructor", err);
    }
    currentApp = this;
}

DBApp::~DBApp()
{
    int err;

    if (imp_) {
        // error prefix is the only piece of memory we manage in the db_env.
        DB_ENV *env = unwrap(this);
        if (env->db_errpfx) {
            delete [] (char *)env->db_errpfx;
            env->db_errpfx = 0;
        }
        err = db_appexit(unwrap(this));
        if (err != 0) {
            DB_DESTRUCTOR_ERROR("DBApp destructor", err);
        }
    }
    //
    // if imp_ is 0, don't bother throwing an error, since the user
    // must have already gotten one from a failure in the constructor.
    //

    if (name_)
        delete [] name_;
}

void DBApp::setErrorModel(DBAppOptions::ErrorModel model)
{
    errorModel_ = model;
}

void DBApp::setName(const char *name)
{
    if (name_)
        delete [] name_;
    name_ = new char[strlen(name)+1];
    strcpy(name_, name);
}

int DBApp::commonCreateTable(DBTable *table_returned,
                             const char *fname,
                             int type,
                             int mode,
                             DBOpenParams params)
{
    DB *newtable;
    int err;
    int flags = params.openFlags_;
    flags |= DB_CREATE;
    flags &= ~DB_RDONLY;        // readonly makes no sense with create
    if ((err = db_open(fname, (DBTYPE)type, flags, mode, unwrap(this),
                       unwrap(&params), &newtable)) != 0) {
        // Note: does not throw exception, only called internally
        return err;
    }
    table_returned->imp_ = wrap(newtable);
    return 0;
}

int DBApp::createBTreeTable(DBBTreeTable *table_returned,
                            const char *fname,
                            int mode,
                            DBOpenBTreeParams params)
{
    int err;
    if ((err = commonCreateTable(table_returned, fname, DB_BTREE,
                                 mode, params)) != 0) {
        DB_ERROR("createBTreeTable", err);
        return err;
    }
    return 0;
}

int DBApp::createHashTable(DBHashTable *table_returned,
                           const char *fname,
                           int mode,
                           DBOpenHashParams params)
{
    int err;
    if ((err = commonCreateTable(table_returned, fname, DB_HASH,
                                 mode, params)) != 0) {
        DB_ERROR("createHashTable", err);
        return err;
    }
    return 0;
}

int DBApp::createRecnoTable(DBRecnoTable *table_returned,
                            const char *fname,
                            int mode,
                            DBOpenRecnoParams params)
{
    int err;
    if ((err = commonCreateTable(table_returned, fname, DB_RECNO,
                                 mode, params)) != 0) {
        DB_ERROR("createRecnoTable", err);
        return err;
    }
    return 0;
}

int DBApp::commonOpenTable(DBTable *table_returned,
                           const char *fname,
                           int type,
                           DBOpenParams params)
{
    DB *newtable;
    int err;
    int flags = params.openFlags_;
    flags &= ~DB_CREATE;        // make sure create is off.
    flags &= ~DB_TRUNCATE;      // truncate makes no sense with open
    if ((err = db_open(fname, (DBTYPE)type, flags, 0,
                       unwrap(this), unwrap(&params), &newtable)) != 0) {
        // Note: does not throw exception, only called internally
        return err;
    }
    table_returned->imp_ = wrap(newtable);
    return 0;
}

int DBApp::openBTreeTable(DBBTreeTable *table_returned,
                          const char *fname,
                          DBOpenBTreeParams params)
{
    int err;
    if ((err = commonOpenTable(table_returned, fname, DB_BTREE, params)) != 0) {
        DB_ERROR("openBTreeTable", err);
        return err;
    }
    return 0;
}

int DBApp::openHashTable(DBHashTable *table_returned,
                         const char *fname,
                         DBOpenHashParams params)
{
    int err;
    if ((err = commonOpenTable(table_returned, fname, DB_HASH, params)) != 0) {
        DB_ERROR("openHashTable", err);
        return err;
    }
    return 0;
}

int DBApp::openRecnoTable(DBRecnoTable *table_returned,
                          const char *fname,
                          DBOpenRecnoParams params)
{
    int err;
    if ((err = commonOpenTable(table_returned, fname, DB_RECNO, params)) != 0) {
        DB_ERROR("openRecno", err);
        return err;
    }
    return 0;
}

int DBApp::openUnknownTable(DBTable *table_returned,
                            const char *fname,
                            DBOpenUnknownParams params)
{
    int err;
    if ((err = commonOpenTable(table_returned, fname, DB_UNKNOWN, params)) != 0) {
        DB_ERROR("openUnknownTable", err);
        return err;
    }
    return 0;
}

DBLockTable DBApp::getLockTable() const
{
    const DB_ENV *env = unwrapConst(this);
    DBLockTable table;
    table.imp_ = wrap(env->lk_info);
    return table;
}

DBLog DBApp::getLog() const
{
    const DB_ENV *env = unwrapConst(this);
    DBLog log;
    log.imp_ = wrap(env->lg_info);
    return log;
}

DBMemPool DBApp::getMemPool() const
{
    const DB_ENV *env = unwrapConst(this);
    DBMemPool mpool;
    mpool.imp_ = wrap(env->mp_info);
    return mpool;
}

DBTransactionManager DBApp::getTransactionManager() const
{
    const DB_ENV *env = unwrapConst(this);
    DBTransactionManager txnmgr;
    txnmgr.imp_ = wrap(env->tx_info);
    return txnmgr;
}

int DBApp::removeLockTableFile(const char *dir, int force)
{
    DB_ENV *env = unwrap(this);
    int err;
    if ((err = lock_unlink(dir, force, env)) != 0) {
        DB_ERROR("DBApp::removeLockTableFile", err);
        return err;
    }
    return 0;
}

int DBApp::removeMemPoolFile(const char *dir, int force)
{
    DB_ENV *env = unwrap(this);
    int err;
    if ((err = memp_unlink(dir, force, env)) != 0) {
        DB_ERROR("DBApp::removeMemPoolFile", err);
        return err;
    }
    return 0;
}

int DBApp::removeTransactionManagerFile(const char *dir, int force)
{
    DB_ENV *env = unwrap(this);
    int err;
    if ((err = txn_unlink(dir, force, env)) != 0) {
        DB_ERROR("DBApp::removeTransactionManagerFile", err);
        return err;
    }
    return 0;
}

int DBApp::removeLoggingFile(const char *dir, int force)
{
    DB_ENV *env = unwrap(this);
    int err;
    if ((err = log_unlink(dir, force, env)) != 0) {
        DB_ERROR("DBApp::removeLoggingFile", err);
        return err;
    }
    return 0;
}

int DBApp::runtimeError(const char *caller, int err, int in_destructor)
{
    DBApp *app = getApp();
    int throwit = (!app ||
                   (app && app->errorModel_ == DBAppOptions::Exception));

    if (throwit && !in_destructor) {
        throw new DBException(caller, err);
    }
    return err;
}

DBApp *DBApp::getApp()
{
    return currentApp;
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBConfig                                //
//                                                                    //
////////////////////////////////////////////////////////////////////////

typedef char *charstar;         // cleans up new syntax

// sequence of name/value pairs stored as an array of "NAME VALUE"
// strings.  Since this has no performance requirements, we reallocate
// on every addition.
//
DBConfig::DBConfig()
:   len_(0)
,   params_(new charstar[1])
{
    params_[0] = 0;
}

DBConfig::~DBConfig()
{
    for (int i=0; i<len_; i++) {
        delete [] params_[i];
    }
    if (params_) {
        delete [] params_;
    }
}

void DBConfig::setParam(const char *name, const char *value, int replace)
{
    // build the new parameter string
    //
    char *newentry = new char[strlen(name) + strlen(value) + 2];
    strcpy(newentry, name);
    strcat(newentry, " ");
    strcat(newentry, value);

    int newindex = -1;
    if (replace)
        newindex = findParamIndex(name, 0);

    if (newindex >= 0) {
        delete params_[newindex];
        params_[newindex] = newentry;
    }
    else {
        // copy all the old entries to a new parameter array.
        //
        char **newparams = new charstar[len_ + 2]; // +1 for new entry, +1 for null
        for (int i=0; i<len_; i++) {
            newparams[i] = params_[i];
        }

        // Add the new entry, and null terminate.
        //
        newparams[len_++] = newentry;
        newparams[len_] = 0;
        newindex = len_;
        delete [] params_;
        params_ = newparams;
    }
}

// Get the "name" parameter from the list.  If there
// are multiple, get the "which" one.
//
const char *DBConfig::getParam(const char *name, int which) const
{
    int index = findParamIndex(name, which);
    if (index < 0)
        return 0;

    return &params_[index][strlen(name)+1];
}

// Get the index of the parameter in the list.  If there
// are multiple, get the "which" one.
//
int DBConfig::findParamIndex(const char *name, int which) const
{
    int namelen = strlen(name);
    for (int i=0; i<len_; i++) {
        if (strncmp(params_[i], name, namelen) == 0 &&
            params_[i][namelen] == ' ') {
            if (which-- <= 0)
                return i;
        }
    }
    return -1;
}

void DBConfig::setLogDirectory(const char *dir)
{
    setParam("DB_LOG_DIR", dir, 1);
}

const char *DBConfig::getLogDirectory() const
{
    return getParam("DB_LOG_DIR");
}

void DBConfig::setTempDirectory(const char *dir)
{
    setParam("DB_TMP_DIR", dir, 1);
}

const char *DBConfig::getTempDirectory() const
{
    return getParam("DB_TMP_DIR");
}

void DBConfig::setDataDirectoryForCreation(const char *dir)
{
    setParam("DB_DATA_DIR", dir, 1);
}

void DBConfig::addDataDirectory(const char *dir)
{
    setParam("DB_DATA_DIR", dir, 0);
}

const char *DBConfig::getDataDirectory(int which) const
{
    return getParam("DB_DATA_DIR", which);
}



////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBAppOptions                            //
//                                                                    //
////////////////////////////////////////////////////////////////////////

// Note: in theory, the db_home and db_*_dir fields will always be zero
// when managed by DBAppOptions.  That's because they are set by
// db_appinit, not by the user, and we make a copy of the db_env used by
// the application.
//

ostream *DBAppOptions::errorStream_ = 0;

DBAppOptions::DBAppOptions()
:   imp_(0)
,   options_(DB_CREATE)
,   errorModel_(DBAppOptions::Exception)
{
    DB_ENV *env = new DB_ENV;
    memset(env, 0, sizeof(DB_ENV));
    imp_ = wrapOptions(env);
}

DBAppOptions::~DBAppOptions()
{
    DB_ENV *env = unwrapOptions(this);
    if (env->db_errpfx)
        delete [] (char *)env->db_errpfx;
    if (env->db_home)
        delete [] env->db_home;
    if (env->db_data_dir) {
        for (int i = 0; env->db_data_dir[i]; i++) {
            delete [] env->db_data_dir[i];
        }
        delete [] env->db_data_dir;
    }
    if (env->db_log_dir)
        delete [] env->db_log_dir;
    if (env->db_tmp_dir)
        delete [] env->db_tmp_dir;
    delete env;
}

DBAppOptions::DBAppOptions(const DBAppOptions &that)
:   imp_(0)
,   options_(that.options_)
,   errorModel_(that.errorModel_)
{
    DB_ENV *env = new DB_ENV;
    imp_ = wrapOptions(env);
    *env = *(unwrapOptionsConst(&that));
    env->db_errpfx = dup(env->db_errpfx);
    env->db_home = dup(env->db_home);
    env->db_data_dir = dup_strarray(env->db_data_dir);
    env->db_log_dir = dup(env->db_log_dir);
    env->db_tmp_dir = dup(env->db_tmp_dir);
}

DBAppOptions &DBAppOptions::operator = (const DBAppOptions &that)
{
    errorModel_ = that.errorModel_;
    options_ = that.options_;
    DB_ENV *env = unwrapOptions(this);
    *env = *(unwrapOptionsConst(&that));
    env->db_errpfx = dup(env->db_errpfx);
    env->db_home = dup(env->db_home);
    env->db_data_dir = dup_strarray(env->db_data_dir);
    env->db_log_dir = dup(env->db_log_dir);
    env->db_tmp_dir = dup(env->db_tmp_dir);
    return *this;
}

void DBAppOptions::setErrorModel(ErrorModel model)
{
    errorModel_ = model;
}

// Note: This actually behaves a bit like a static function,
// since DB_ENV.db_errcall has no information about which
// db_env triggered the call.  A user that has multiple DB_ENVs
// will simply not be able to have different streams for each one.
//
void DBAppOptions::setErrorStream(class ostream *stream)
{
    errorStream_ = stream;

    DB_ENV *env = unwrapOptions(this);
    env->db_errcall = StreamErrorFunction;
}

ostream *DBAppOptions::getErrorStream() const
{
    return errorStream_;
}

void DBAppOptions::StreamErrorFunction(const char *prefix, char *message)
{
    if (errorStream_) {
        if (prefix) {
            (*errorStream_) << prefix << ": ";
        }
        if (message) {
            (*errorStream_) << message;
        }
        (*errorStream_) << "\n";
    }
}

// The error prefix field is allocated and deallocated as needed.
//
void DBAppOptions::setErrorPrefix(const char * value)
{
    DB_ENV *env = unwrapOptions(this);
    if (env->db_errpfx)
        delete [] (char *)env->db_errpfx;
    env->db_errpfx = dup(value);
}

const char * DBAppOptions::getErrorPrefix() const
{
    const DB_ENV *env = unwrapOptionsConst(this);
    return (env->db_errpfx);
}

DB_FLAG_METHODS(DBAppOptions, options_, UseLocking, DB_INIT_LOCK)
DB_FLAG_METHODS(DBAppOptions, options_, UseLogging, DB_INIT_LOG)
DB_FLAG_METHODS(DBAppOptions, options_, UseMemPools, DB_INIT_MPOOL)
DB_FLAG_METHODS(DBAppOptions, options_, UseTransactions, DB_INIT_TXN)
DB_FLAG_METHODS(DBAppOptions, options_, CreateFiles, DB_CREATE)
DB_FLAG_METHODS(DBAppOptions, options_, UseEnvironment, DB_USE_ENVIRON)
DB_FLAG_METHODS(DBAppOptions, options_,
                 UseEnvironmentForRoot, DB_USE_ENVIRON_ROOT)
DB_FLAG_METHODS(DBAppOptions, options_, PrivateMemPool, DB_MPOOL_PRIVATE)


#define DB_OPTIONS_ACCESS(_type, _cxx_name, _field, _optional_dup) \
                                                     \
void DBAppOptions::set##_cxx_name(_type value)       \
{                                                    \
    DB_ENV *env = unwrapOptions(this);               \
    env->_field = _optional_dup(value);              \
}                                                    \
                                                     \
_type DBAppOptions::get##_cxx_name() const           \
{                                                    \
    const DB_ENV *env = unwrapOptionsConst(this);    \
    return (env->_field);                            \
}

DB_OPTIONS_ACCESS(DBAppOptions::YieldFunction, YieldFunction, db_yield,
                  (DBAppOptions::YieldFunction))
DB_OPTIONS_ACCESS(u_int8_t *, LockingConflictMatrix,
                   lk_conflicts, (u_int8_t*))
DB_OPTIONS_ACCESS(int, LockModes, lk_modes, (int))
DB_OPTIONS_ACCESS(unsigned int, MaxLocks, lk_max, (unsigned int))
DB_OPTIONS_ACCESS(u_int32_t, MaxLogFileSize, lg_max, (u_int32_t))
DB_OPTIONS_ACCESS(size_t, MaxMemPoolMmapSize, mp_mmapsize, (size_t))
DB_OPTIONS_ACCESS(size_t, MemPoolCacheSize, mp_size, (size_t))
DB_OPTIONS_ACCESS(unsigned int, MaxTransactions, tx_max, (unsigned int))
DB_OPTIONS_ACCESS(DBAppOptions::TransactionRecoveryFunction,
                   TransactionRecoveryFunction, tx_recover,
                   (DBAppOptions::TransactionRecoveryFunction))
DB_OPTIONS_ACCESS(FILE *, ErrorFile, db_errfile, (FILE *))
DB_OPTIONS_ACCESS(DBAppOptions::ErrorFunction, ErrorFunction,
                   db_errcall, (DBAppOptions::ErrorFunction))
