/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)cxx_mpool.cpp	10.3 (Sleepycat) 7/20/97";
#endif /* not lint */

#include "db_cxx.h"
#include "cxx_int.h"
#include <errno.h>

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBMemPoolFile                           //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBMemPoolFile::DBMemPoolFile(DBMemPool &mp, const char *file,
                               int ftype, int flags, int mode,
                               size_t pagesize, int lsn_offset,
                               DBThunk &pgcookie)
:   imp_(0)
{
    int err;

    DB_MPOOLFILE *mpf;
    if ((err = memp_fopen(unwrap(&mp), file, ftype, flags, mode, pagesize,
                          lsn_offset, unwrap(&pgcookie), &mpf)) != 0) {
        DB_ERROR("DBMemPoolFile::DBMemPoolFile", err);
    }
    imp_ = wrap(mpf);
}

DBMemPoolFile::~DBMemPoolFile()
{
    // Note: could do automatic close, but only if we reference count
}

int DBMemPoolFile::close()
{
    DB_MPOOLFILE *mpf = unwrap(this);
    int err = 0;
    if (!mpf) {
        err = EINVAL;
    }
    else if ((err = memp_fclose(mpf)) != 0) {
        DB_ERROR("DBMemPoolFile::close", err);
    }
    return err;
}

int DBMemPoolFile::get(db_pgno_t *pgnoaddr, unsigned long flags, void *pagep)
{
    DB_MPOOLFILE *mpf = unwrap(this);
    int err = 0;
    if (!mpf) {
        err = EINVAL;
    }
    else if ((err = memp_fget(mpf, pgnoaddr, flags, pagep)) != 0) {
        DB_ERROR("DBMemPoolFile::get", err);
    }
    return err;
}

int DBMemPoolFile::put(void *pgaddr, unsigned long flags)
{
    DB_MPOOLFILE *mpf = unwrap(this);
    int err = 0;
    if (!mpf) {
        err = EINVAL;
    }
    else if ((err = memp_fput(mpf, pgaddr, flags)) != 0) {
        DB_ERROR("DBMemPoolFile::put", err);
    }
    return err;
}

int DBMemPoolFile::set(void *pgaddr, unsigned long flags)
{
    DB_MPOOLFILE *mpf = unwrap(this);
    int err = 0;
    if (!mpf) {
        err = EINVAL;
    }
    else if ((err = memp_fset(mpf, pgaddr, flags)) != 0) {
        DB_ERROR("DBMemPoolFile::set", err);
    }
    return err;
}

int DBMemPoolFile::sync()
{
    DB_MPOOLFILE *mpf = unwrap(this);
    int err = 0;
    if (!mpf) {
        err = EINVAL;
    }
    else if ((err = memp_fsync(mpf)) != 0) {
        DB_ERROR("DBMemPoolFile::sync", err);
    }
    return err;
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBMemPool                               //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBMemPool::DBMemPool()
:   imp_(0)
{
}

DBMemPool::DBMemPool(const char *dir, int flags, int mode, DBApp &app)
{
    DB_MPOOL *mpool = 0;
    int err = 0;
    if ((err = memp_open(dir, flags, mode, unwrap(&app), &mpool)) != 0) {
        DB_ERROR("DBMemPool::DBMemPool", err);
    }
    imp_ = wrap(mpool);
}

DBMemPool::~DBMemPool()
{
    // Note: could do automatic close, but only if we reference count
}

int DBMemPool::close()
{
    DB_MPOOL *mpool = unwrap(this);
    int err;

    if (!mpool) {
        return EINVAL;        // already closed
    }

    if ((err = memp_close(mpool)) != 0) {
        DB_ERROR("DBMemPool::close", err);
    }
    imp_ = 0;
    return err;
}

int DBMemPool::identify(int ftype,
                     int (*pgin)(db_pgno_t pgno, void *pgaddr, DBT *pgcookie),
                     int (*pgout)(db_pgno_t pgno, void *pgaddr, DBT *pgcookie))
{
    DB_MPOOL *mpool = unwrap(this);
    int err = 0;
    if (!mpool) {
        err = EINVAL;
    }
    else if ((err = memp_register(mpool, ftype, pgin, pgout)) != 0) {
        DB_ERROR("DBMemPool::identify", err);
    }
    return err;
}

int DBMemPool::stat(DB_MPOOL_STAT **gsp, DB_MPOOL_FSTAT ***fsp,
                     void *(*alternate_malloc)(size_t))
{
    DB_MPOOL *mpool = unwrap(this);
    int err = 0;
    if (!mpool) {
        err = EINVAL;
    }
    else if ((err = memp_stat(mpool, gsp, fsp, alternate_malloc)) != 0) {
        DB_ERROR("DBMemPool::stat", err);
    }
    return err;
}

int DBMemPool::sync(DBSequenceNumber &sn)
{
    DB_MPOOL *mpool = unwrap(this);
    int err = 0;
    if (!mpool) {
        err = EINVAL;
    }
    else if ((err = memp_sync(mpool, unwrap(&sn))) != 0) {
        DB_ERROR("DBMemPool::sync", err);
    }
    return err;
}

int DBMemPool::isValid() const
{
    return (unwrapConst(this) != 0);
}
