/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1996, 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1997\n\
	Sleepycat Software Inc.  All rights reserved.\n";
static const char sccsid[] = "@(#)db_load.c	10.7 (Sleepycat) 7/2/97";
#endif

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "db_int.h"
#include "clib_ext.h"

void	configure __P((char *));
DB_ENV *db_init __P((char *));
int	dbt_rdump __P((DBT *));
int	dbt_rprint __P((DBT *));
int	digitize __P((int));
void	rheader __P((DBTYPE *, int *, DB_INFO *));
void	usage __P((void));

char *progname = "db_load";		/* Program name. */

long	c_db_lorder;			/* configure: byte order. */
int	c_db_lorder_set;		/* configure: byte order set. */
long	c_db_pagesize;			/* configure: page size. */
int	c_db_pagesize_set;		/* configure: page size set. */

int
main(argc, argv)
	int argc;
	char *argv[];
{
	extern char *optarg;
	extern int optind;
	DB *dbp;
	DBT key, data;
	DBTYPE argtype, headertype;
	DB_ENV *dbenv;
	DB_INFO dbinfo;
	int ch, pflag;
	char *home;

	home = NULL;
	argtype = DB_UNKNOWN;
	while ((ch = getopt(argc, argv, "c:f:h:t:")) != EOF)
		switch (ch) {
		case 'c':
			configure(optarg);
			break;
		case 'f':
			if (freopen(optarg, "r", stdin) == NULL)
				err(1, "%s", optarg);
			break;
		case 'h':
			home = optarg;
			break;
		case 't':
			if (strcmp(optarg, "btree") == 0) {
				argtype = DB_BTREE;
				break;
			}
			if (strcmp(optarg, "hash") == 0) {
				argtype = DB_HASH;
				break;
			}
			usage();
			/* NOTREACHED */
		case '?':
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();

	/* Initialize the environment. */
	dbenv = db_init(home);

	/* Read the header. */
	memset(&dbinfo, 0, sizeof(DB_INFO));
	rheader(&headertype, &pflag, &dbinfo);

	/* Apply command-line configuration changes. */
	if (c_db_lorder_set)
		dbinfo.db_lorder = c_db_lorder;
	if (c_db_pagesize_set)
		dbinfo.db_pagesize = c_db_pagesize;

	/* Conversion to/from recno is prohibited. */
	if (argtype != DB_UNKNOWN) {
		if (headertype == DB_RECNO)
			errx(1, "databases of type recno may not be converted");
		headertype = argtype;
	}

	/* Open the DB file. */
	if ((errno = db_open(argv[0], headertype, DB_CREATE | DB_TRUNCATE,
	    S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH,
	    dbenv, &dbinfo, &dbp)) != 0)
		err(1, "%s", argv[0]);

	/* Initialize the key/data pair. */
	memset(&key, 0, sizeof(DBT));
	if ((key.data = (void *)malloc(key.ulen = 1024)) == NULL) {
		errno = ENOMEM;
		err(1, NULL);
	}
	memset(&data, 0, sizeof(DBT));
	if ((data.data = (void *)malloc(data.ulen = 1024)) == NULL) {
		errno = ENOMEM;
		err(1, NULL);
	}

	/* Get each key/data pair and add them to the database. */
	for (;;) {
		if (pflag) {
			if (dbt_rprint(&key))
				break;
			if (dbt_rprint(&data))
				goto fmt;
		} else {
			if (dbt_rdump(&key))
				break;
			if (dbt_rdump(&data))
fmt:				err(1, "odd number of key/data pairs");
		}
		if ((errno = dbp->put(dbp, NULL, &key, &data, 0)) != 0)
			err(1, "%s", argv[0]);
	}
	if ((errno = dbp->close(dbp, 0)) != 0)
		err(1, "%s", argv[0]);
	return (0);
}

/*
 * db_init --
 *	Initialize the environment.
 */
DB_ENV *
db_init(home)
	char *home;
{
	DB_ENV *dbenv;

	if ((dbenv = (DB_ENV *)calloc(sizeof(DB_ENV), 1)) == NULL) {
		errno = ENOMEM;
		err(1, NULL);
	}
	dbenv->db_errfile = stderr;
	dbenv->db_errpfx = progname;

	if ((errno =
	    db_appinit(home, NULL, dbenv, DB_CREATE | DB_USE_ENVIRON)) != 0)
		err(1, "db_appinit");
	return (dbenv);
}

/*
 * configure --
 *	Handle command-line configuration options.
 */
void
configure(arg)
	char *arg;
{
	char *eqp;

	if ((eqp = strchr(arg, '=')) == NULL)
		errx(1, "command-line configuration use name=value format");
	*eqp++ = '\0';

#undef	CONFIG
#define	CONFIG(name, value, flag)					\
	if (strcmp(arg, name) == 0) {					\
		get_long(eqp, 1, LONG_MAX, &value);			\
		flag = 1;						\
		return;							\
	}
	CONFIG("db_lorder", c_db_lorder, c_db_lorder_set);
	CONFIG("db_pagesize", c_db_pagesize, c_db_pagesize_set);

	errx(1, "unknown command-line configuration keyword");
}

/*
 * rheader --
 *	Read the header message.
 */
void
rheader(dbtypep, pflagp, dbinfop)
	DBTYPE *dbtypep;
	int *pflagp;
	DB_INFO *dbinfop;
{
	long lineno, val;
	char name[256], value[256];

	*dbtypep = DB_UNKNOWN;
	*pflagp = 0;

/*
 * XXX
 * There's no bounds checking on the value.
 */
#undef	FIELD
#define	FIELD(keyword, fname)						\
	if (strcmp(name, keyword) == 0) {				\
		get_long(value, 1, LONG_MAX, &val);			\
		dbinfop->fname = val;					\
		continue;						\
	}
#undef	FLAG
#define	FLAG(keyword, flag)						\
	if (strcmp(name, keyword) == 0) {				\
		if (value)						\
			dbinfop->flags |= (flag);			\
		else							\
			dbinfop->flags &= ~(flag);			\
		continue;						\
	}
	for (lineno = 1;; ++lineno) {
		if (fscanf(stdin, "%[^=]=%s\n", name, value) != 2)
			errx(1, "line %lu: unexpected line", lineno);
		if (strcmp(name, "HEADER") == 0)
			break;

		FIELD("bt_maxkey", bt_maxkey);
		FIELD("bt_minkey", bt_minkey);
		FIELD("db_lorder", db_lorder);
		FIELD("db_pagesize", db_pagesize);
		FLAG("duplicates", DB_DUP);
		FIELD("h_ffactor", h_ffactor);
		FIELD("h_nelem", h_nelem);
		FIELD("re_delim", re_delim);
		FIELD("re_len", re_len);
		FIELD("re_pad", re_pad);

		if (strcmp(name, "format") == 0) {
			if (strcmp(value, "bytevalue") == 0) {
				*pflagp = 0;
				continue;
			}
			if (strcmp(value, "print") == 0) {
				*pflagp = 1;
				continue;
			}
			errx(1, "line %d: unknown format", lineno);
		}
		if (strcmp(name, "type") == 0) {
			if (strcmp(value, "btree") == 0) {
				*dbtypep = DB_BTREE;
				continue;
			}
			if (strcmp(value, "hash") == 0) {
				*dbtypep = DB_HASH;
				continue;
			}
			if (strcmp(value, "recno") == 0) {
				*dbtypep = DB_RECNO;
				continue;
			}
			errx(1, "line %d: unknown type", lineno);
		}
	}
}

/*
 * dbt_rprint --
 *	Read a printable line into a DBT structure.
 */
int
dbt_rprint(dbtp)
	DBT *dbtp;
{
	u_int32_t len;
	u_int8_t *p;
	int c1, c2, escape;

	escape = 0;
	for (p = dbtp->data, len = 0; (c1 = getchar()) != '\n';) {
		if (c1 == EOF) {
			if (len == 0)
				return (1);
			err(1, "unexpected end of key/data pair");
		}
		if (escape) {
			if (c1 != '\\') {
				if ((c2 = getchar()) == EOF)
					err(1,
					    "unexpected end of key/data pair");
				c1 = digitize(c1) << 4 | digitize(c2);
			}
			escape = 0;
		} else
			if (c1 == '\\') {
				escape = 1;
				continue;
			}
		if (++len >= dbtp->ulen - 10) {
			dbtp->ulen *= 2;
			if ((dbtp->data =
			    (void *)realloc(dbtp->data, dbtp->ulen)) == NULL) {
				errno = ENOMEM;
				err(1, NULL);
			}
			p = (u_int8_t *)dbtp->data + len;
		}
		*p++ = c1;
	}
	dbtp->size = len;
	return (0);
}

/*
 * digitize --
 *	Convert a character to an integer.
 */
int
digitize(c)
	int c;
{
	switch (c) {			/* Don't depend on ASCII ordering. */
	case '0': return (0);
	case '1': return (1);
	case '2': return (2);
	case '3': return (3);
	case '4': return (4);
	case '5': return (5);
	case '6': return (6);
	case '7': return (7);
	case '8': return (8);
	case '9': return (9);
	case 'a': return (10);
	case 'b': return (11);
	case 'c': return (12);
	case 'd': return (13);
	case 'e': return (14);
	case 'f': return (15);
	}

	err(1, "unexpected hexadecimal value");
	/* NOTREACHED */

	return (0);
}

/*
 * dbt_rdump --
 *	Read a byte dump line into a DBT structure.
 */
int
dbt_rdump(dbtp)
	DBT *dbtp;
{
	u_int32_t len;
	u_int8_t *p;
	int c1, c2;

	for (p = dbtp->data, len = 0; (c1 = getchar()) != '\n';) {
		if (c1 == EOF) {
			if (len == 0)
				return (1);
			err(1, "unexpected end of key/data pair");
		}
		if ((c2 = getchar()) == EOF)
			err(1, "unexpected end of key/data pair");
		if (++len >= dbtp->ulen - 10) {
			dbtp->ulen *= 2;
			if ((dbtp->data =
			    (void *)realloc(dbtp->data, dbtp->ulen)) == NULL) {
				errno = ENOMEM;
				err(1, NULL);
			}
			p = (u_int8_t *)dbtp->data + len;
		}
		*p++ = digitize(c1) << 4 | digitize(c2);
	}
	dbtp->size = len;
	return (0);
}

/*
 * usage --
 *	Display the usage message.
 */
void
usage()
{
	(void)fprintf(stderr,
"usage: db_load [-c name=value] [-f file] [-h home] [-t btree | hash] db_file\n");
	exit(1);
}
