/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1996, 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1997\n\
	Sleepycat Software Inc.  All rights reserved.\n";
static const char sccsid[] = "@(#)ex_mpool.c	10.6 (Sleepycat) 6/19/97";
#endif

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#endif

#include <db.h>

#define	MPOOL	"mpool"

void	init(char *, int, int);
void	run(DB_ENV *, int, int, int);
void	usage(void);

char *progname = "ex_mpool";				/* Program name. */

int
main(argc, argv)
	int argc;
	char *argv[];
{
	extern char *optarg;
	extern int optind;
	DB_ENV dbenv;
	int cachesize, ch, hits, npages, pagesize;

	cachesize = 20 * 1024;
	hits = 1000;
	npages = 50;
	pagesize = 1024;
	while ((ch = getopt(argc, argv, "c:h:n:p:")) != EOF)
		switch (ch) {
		case 'c':
			if ((cachesize = atoi(optarg)) < 20 * 1024)
				usage();
			break;
		case 'h':
			if ((hits = atoi(optarg)) <= 0)
				usage();
		case 'n':
			if ((npages = atoi(optarg)) <= 0)
				usage();
			break;
		case 'p':
			if ((pagesize = atoi(optarg)) <= 0)
				usage();
			break;
		case '?':
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	/* Initialize the file. */
	init(MPOOL, pagesize, npages);

	/*
	 * Initialize the environment.
	 *
	 * Output error messages to stderr.
	 * Specify the cachesize.
	 * The memory pool is private.
	 */
	memset(&dbenv, 0, sizeof(dbenv));
	dbenv.db_errfile = stderr;
	dbenv.db_errpfx = "ex_mpool";
	dbenv.mp_size = cachesize;
	dbenv.flags |= DB_MPOOL_PRIVATE;

	/* Get the pages. */
	run(&dbenv, hits, pagesize, npages);

	return (0);
}

/*
 * init --
 *	Create a backing file.
 */
void
init(file, pagesize, npages)
	char *file;
	int pagesize, npages;
{
	int cnt, fd;
	char *p;

	/*
	 * Create a file with the right number of pages, and store a page
	 * number on each page.
	 */
	if ((fd = open(file, O_CREAT | O_RDWR | O_TRUNC, 0666)) < 0) {
		fprintf(stderr, "ex_mpool: %s: %s\n", file, strerror(errno));
		exit (1);
	}
	if ((p = malloc(pagesize)) == NULL) {
		fprintf(stderr, "ex_mpool: %s\n", strerror(ENOMEM));
		exit (1);
	}

	/* The pages are numbered from 0. */
	for (cnt = 0; cnt <= npages; ++cnt) {
		*(int *)p = cnt;
		if (write(fd, p, pagesize) != pagesize) {
			fprintf(stderr,
			    "ex_mpool: %s: %s\n", file, strerror(errno));
			exit (1);
		}
	}
	free(p);
}

/*
 * run --
 *	Get a set of pages.
 */
void
run(dbenv, hits, pagesize, npages)
	DB_ENV *dbenv;
	int hits, pagesize, npages;
{
	DB_MPOOL *dbmp;
	DB_MPOOLFILE *dbmfp;
	db_pgno_t pageno;
	int cnt;
	void *p;

	/* Open a memory pool. */
	if ((errno = memp_open(NULL, DB_CREATE, 0666, dbenv, &dbmp)) != 0) {
		fprintf(stderr, "ex_mpool: %s\n", strerror(errno));
		exit (1);
	}
	/* Open the file in the pool. */
	if ((errno = memp_fopen(dbmp,
	    MPOOL, 0, 0, 0, pagesize, 0, NULL, &dbmfp)) != 0) {
		fprintf(stderr, "ex_mpool: %s: %s\n", MPOOL, strerror(errno));
		exit (1);
	}

	srand((u_int)time(NULL));
	for (cnt = 0; cnt < hits; ++cnt) {
		pageno = (rand() % npages) + 1;
		if ((errno = memp_fget(dbmfp, &pageno, 0, &p)) != 0) {
			fprintf(stderr,
			    "ex_mpool: unable to retrieve page %lu: %s\n",
			    (u_long)pageno, strerror(errno));
			exit (1);
		}
		if (*(int *)p != pageno) {
			fprintf(stderr,
			    "ex_mpool: wrong page retrieved (%lu != %d)\n",
			    (u_long)pageno, *(int *)p);
			exit (1);
		}
		if ((errno = memp_fput(dbmfp, p, 0)) != 0) {
			fprintf(stderr,
			    "ex_mpool: unable to return page %lu: %s\n",
			    (u_long)pageno, strerror(errno));
			exit (1);
		}
	}

	/* Close the pool. */
	if ((errno = memp_close(dbmp)) != 0) {
		fprintf(stderr, "ex_mpool: %s\n", strerror(errno));
		exit (1);
	}
}

void
usage()
{
	(void)fprintf(stderr,
    "usage: ex_mpool [-c cachesize] [-h hits] [-n npages] [-p pagesize]\n");
	exit(1);
}
