/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1997\n\
	Sleepycat Software Inc.  All rights reserved.\n";
static const char sccsid[] = "@(#)TpcbExample.cpp	10.2 (Sleepycat) 7/20/97";
#endif

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>

#include <errno.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif

#include <db_cxx.h>

typedef enum { ACCOUNT, BRANCH, TELLER } FTYPE;

void errExit(int err, const char *);  // show err as errno and exit

void	  invarg(int, char *);
u_int32_t random_id(FTYPE, u_int32_t, u_int32_t, u_int32_t);
u_int32_t random_int(u_int32_t, u_int32_t);
static void	  usage(void);

int verbose;
char *progname = "TpcbExample";                            // Program name.

class TpcbExample : public DBApp
{
public:
    void populate(int, int, int, int);
    void run(int, int, int, int);
    int txn(DBTransactionManager,
            DBTable, DBTable, DBTable, DBTable,
            int, int, int);
    void populateHistory(DBTable &, unsigned int, u_int32_t,
                         u_int32_t, u_int32_t);
    void populateTable(DBTable &, u_int32_t, u_int32_t, unsigned int, char *);

    TpcbExample(const char *home, DBAppOptions &options);

private:
    static const char FileName[];

    // no need for copy and assignment
    TpcbExample(const TpcbExample &);
    operator = (const TpcbExample &);
};

//
// This program implements a basic TPC/B driver program.  To create the
// TPC/B database, run with the -i (init) flag.  The number of records
// with which to populate the account, history, branch, and teller tables
// is specified by the a, s, b, and t flags respectively.  To run a TPC/B
// test, use the n flag to indicate a number of transactions to run (note
// that you can run many of these processes in parallel to simulate a
// multiuser test run).
//
// XXX Margo, check these ratios and record sizes.
//
#define TELLERS_PER_BRANCH      100
#define ACCOUNTS_PER_TELLER     1000

#define ACCOUNTS 1000000
#define	BRANCHES 10
#define TELLERS 1000
#define HISTORY 1000000
#define HISTORY_LEN 100
#define RECLEN 100
#define BEGID   1000000

struct Defrec {
    u_int32_t   id;
    u_int32_t   balance;
    u_int8_t    pad[RECLEN - sizeof(u_int32_t) - sizeof(u_int32_t)];
};

struct Histrec {
    u_int32_t   aid;
    u_int32_t   bid;
    u_int32_t   tid;
    u_int32_t   amount;
    u_int8_t    pad[RECLEN - 4 * sizeof(u_int32_t)];
};

int
main(int argc, char *argv[])
{
    unsigned long seed;
    int accounts, branches, tellers, history;
    int ch, iflag, mpool, ntxns;
    char *home, *endarg;

    home = NULL;
    accounts = branches = history = tellers = 0;
    mpool = ntxns = 0;
    verbose = 0;
    iflag = 0;
    seed = (unsigned long)getpid();

    for (int i = 1; i < argc; ++i)
    {

        if (strcmp(argv[i], "-a") == 0) {
            // Number of account records
            if ((accounts = atoi(argv[++i])) <= 0)
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-b") == 0) {
            // Number of branch records
            if ((branches = atoi(argv[++i])) <= 0)
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-h") == 0) {
            // DB  home.
            home = argv[++i];
        }
        else if (strcmp(argv[i], "-i") == 0) {
            // Initialize the test.
            iflag = 1;
        }
        else if (strcmp(argv[i], "-m") == 0) {
            // Bytes in buffer pool
            if ((mpool = atoi(argv[++i])) <= 0)
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-n") == 0) {
            // Number of transactions
            if ((ntxns = atoi(argv[++i])) <= 0)
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-S") == 0) {
            // Random number seed.
            seed = strtoul(argv[++i], &endarg, 0);
            if (*endarg != '\0')
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-s") == 0) {
            // Number of history records
            if ((history = atoi(argv[++i])) <= 0)
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-t") == 0) {
            // Number of teller records
            if ((tellers = atoi(argv[++i])) <= 0)
                invarg(ch, argv[i]);
        }
        else if (strcmp(argv[i], "-v") == 0) {
            // Verbose option.
            verbose = 1;
        }
        else
        {
            usage();
        }
    }

    srand((unsigned int)seed);

    accounts = accounts == 0 ? ACCOUNTS : accounts;
    branches = branches == 0 ? BRANCHES : branches;
    tellers = tellers == 0 ? TELLERS : tellers;
    history = history == 0 ? HISTORY : history;

    if (verbose)
        cout << (long)accounts << " Accounts "
             << (long)branches << " Branches "
             << (long)tellers << " Tellers "
             << (long)history << " History\n";

    // Declaring and setting options does not need
    // to be done in a try block, as it will never
    // raise an exception.
    //
    DBAppOptions options;

    options.setErrorStream(&cerr);
    options.setErrorPrefix("tpcb");
    if (mpool == 0) {
        mpool = 4 * 1024 * 1024;
    }
    options.setMemPoolCacheSize(mpool);
    options.setCreateFiles(1);
    options.setUseMemPools(1);

    if (!iflag) {
        options.setUseTransactions(1);
        options.setUseLogging(1);
        options.setUseLocking(1);
    }

    try {
        // Initialize the database environment.
        // Must be done in within a try block, unless you
        // change the error model in the options.
        //
        TpcbExample app(home, options);

        if (iflag && ntxns != 0) {
            cerr << "specify only one of -i and -n\n";
            exit(1);
        }
        if (iflag)
            app.populate(accounts, branches, history, tellers);
        if (ntxns != 0)
            app.run(ntxns, accounts, branches, tellers);

        return 0;
    }
    catch (DBException &dbe)
    {
        cerr << "TpcbExample: " << dbe.what() << "\n";
        return 1;
    }
}

void
invarg(int arg, char *str)
{
    cerr << "TpcbExample: invalid argument for -"
         << (char)arg << ": " << str << "\n";
    exit(1);
}

static void
usage()
{
    cerr << "usage: TpcbExample "
         << "[-iv] [-a accounts] [-b branches] [-h home] [-m mpool_size] "
         << "[-n transactions ] [-S seed] [-s history] [-t tellers] \n";
    exit(1);
}

TpcbExample::TpcbExample(const char *home, DBAppOptions &options)
:   DBApp(home, options)
{
}

//
// Initialize the database to the specified number of accounts, branches,
// history records, and tellers.
//
// Note: num_h was unused in the original ex_tpcb.c example.
//
void
TpcbExample::populate(int num_a, int num_b, int /*num_h*/, int num_t)
{
    DBHashTable dbp;

    u_int32_t balance, idnum;
    u_int32_t end_anum, end_bnum, end_tnum;
    u_int32_t start_anum, start_bnum, start_tnum;

    idnum = BEGID;
    balance = 500000;

    DBOpenHashParams params;
    params.setNumberOfElements((unsigned int)num_a);
    params.setTruncate(1);

    int err;
    if ((err = createHashTable(&dbp, "account", 0644, params)) != 0)
        errExit(err, "tpcb: Open of account file failed");

    start_anum = idnum;
    populateTable(dbp, idnum, balance, params.getNumberOfElements(), "account");
    idnum += params.getNumberOfElements();
    end_anum = idnum - 1;
    if ((err = dbp.close(0)) != 0)
        errExit(err, "tpcb: Close of account file failed");

    if (verbose)
        cout << "Populated accounts: "
             << (long)start_anum << " - " << (long)end_anum << "\n";


    //
    // Since the number of branches is very small, we want to use very
    // small pages and only 1 key per page.  This is the poor-man's way
    // of getting key locking instead of page locking.
    //
    params.setNumberOfElements((unsigned int)num_b);
    params.setFillFactor(1);
    params.setPageSize(512);

    if ((err = createHashTable(&dbp, "branch", 0644, params)) != 0)
        errExit(err, "tpcb: Create of branch file failed");
    start_bnum = idnum;
    populateTable(dbp, idnum, balance, params.getNumberOfElements(), "branch");
    idnum += params.getNumberOfElements();
    end_bnum = idnum - 1;
    if ((err = dbp.close(0)) != 0)
        errExit(err, "tpcb: Close of branch file failed");

    if (verbose)
        cout << "Populated branches: "
             << (long)start_bnum << " - " << (long)end_bnum << "\n";


    //
    // In the case of tellers, we also want small pages, but we'll let
    // the fill factor dynamically adjust itself.
    //
    params.setNumberOfElements((unsigned int)num_t);
    params.setFillFactor(0);
    params.setPageSize(512);
    if ((err = createHashTable(&dbp, "teller", 0644, params)) != 0)
        errExit(err, "tpcb: Create of teller file failed");

    start_tnum = idnum;
    populateTable(dbp, idnum, balance, params.getNumberOfElements(), "teller");
    idnum += params.getNumberOfElements();
    end_tnum = idnum - 1;
    if ((err = dbp.close(0)) != 0)
        errExit(err, "tpcb: Close of teller file failed");

    if (verbose)
        cout << "Populated tellers: "
             << (long)start_tnum << " - " << (long)end_tnum << "\n";

    DBRecnoTable histTable;
    DBOpenRecnoParams histParams;
    histParams.setFixedLength(HISTORY_LEN);
    if ((err = createRecnoTable(&histTable, "history", 0644, histParams)) != 0)
        errExit(err, "tpcb: Create of history file failed");

    populateHistory(histTable, num_t, num_a, num_b, num_t);
    if ((err = histTable.close(0)) != 0)
        errExit(err, "tpcb: Close of history file failed");
}

void
TpcbExample::populateTable(DBTable &dbp,
                           u_int32_t start_id, u_int32_t balance,
                           unsigned int nrecs, char *msg)
{
    DBThunk kdbt, ddbt;
    Defrec drec;
    unsigned int i;

    kdbt.setData(&drec.id, sizeof(u_int32_t));
    ddbt.setData(&drec, sizeof(drec));
    memset(&drec.pad[0], 1, sizeof(drec.pad));

    for (i = 0; i < nrecs; i++) {
        drec.id = start_id + (u_int32_t)i;
        drec.balance = balance;
        int err;
        if ((err = dbp.put(kdbt, ddbt)) != 0) {
            cerr << "tpcb: Failure initializing " << msg << " file: "
                 << strerror(err) << "\n";
            exit(1);
        }
    }
}

void
TpcbExample::populateHistory(DBTable &dbp, unsigned int nrecs,
                             u_int32_t anum, u_int32_t bnum, u_int32_t tnum)
{
    DBThunk kdbt, ddbt;
    Histrec hrec;
    u_int32_t i;

    kdbt.setData(&i, sizeof(u_int32_t));
    ddbt.setData(&hrec, sizeof(hrec));

    memset(&hrec.pad[0], 1, sizeof(hrec.pad));
    hrec.amount = 10;

    for (i = 1; i <= nrecs; i++) {
        hrec.aid = random_id(ACCOUNT, anum, bnum, tnum);
        hrec.bid = random_id(BRANCH, anum, bnum, tnum);
        hrec.tid = random_id(TELLER, anum, bnum, tnum);
        int err;
        if ((err = dbp.put(kdbt, ddbt)) != 0)
            errExit(err, "tpcb: Failure initializing history file");
    }
}

u_int32_t
random_int(u_int32_t lo, u_int32_t hi)
{
    u_int32_t ret;
    int t;

    t = rand();
    ret = (u_int32_t)(((double)t / ((double)(RAND_MAX) + 1)) *
                      (hi - lo + 1));
    ret += lo;
    return (ret);
}

u_int32_t
random_id(FTYPE type, u_int32_t accounts, u_int32_t branches, u_int32_t tellers)
{
    u_int32_t min, max, num;

    max = min = BEGID;
    num = accounts;
    switch(type) {
    case TELLER:
        min += branches;
        num = tellers;
        // Fallthrough
    case BRANCH:
        if (type == BRANCH)
            num = branches;
        min += accounts;
        // Fallthrough
    case ACCOUNT:
        max = min + num - 1;
    }
    return (random_int(min, max));
}

void
TpcbExample::run(int n, int accounts, int branches, int tellers)
{
    DBTable adb, bdb, hdb, tdb;
    int failed, ret, txns;

    //
    // Open the database files.
    //
    int err;
    if ((err = openUnknownTable(&adb, "account")) != 0)
        errExit(err, "tpcb: Open of account file failed");
    if ((err = openUnknownTable(&bdb, "branch")) != 0)
        errExit(err, "tpcb: Open of branch file failed");
    if ((err = openUnknownTable(&tdb, "teller")) != 0)
        errExit(err, "tpcb: Open of teller file failed");
    if ((err = openUnknownTable(&hdb, "history")) != 0)
        errExit(err, "tpcb: Open of history file failed");

    // Change the error model so that errors are returned instead
    // of throwing exceptions.  txn() will handle error returns.
    //
    setErrorModel(DBAppOptions::ErrorReturn);

    txns = failed = 0;
    while (n-- >= 0) {
        txns++;
        ret = txn(getTransactionManager(),
                  adb, bdb, tdb, hdb, accounts, branches, tellers);
        if (ret != 0)
            failed++;
    }

    adb.close(0);
    bdb.close(0);
    tdb.close(0);
    hdb.close(0);

    cout << (long)txns << " transactions begun "
         << (long)failed << " failed\n";
}

//
// XXX Figure out the appropriate way to pick out IDs.
//
int
TpcbExample::txn(DBTransactionManager txmgr,
                 DBTable adb, DBTable bdb, DBTable tdb, DBTable hdb,
                 int anum, int bnum, int tnum)
{
    DBIterator acurs, bcurs, hcurs, tcurs;
    DBThunk d_dbt, d_histdbt, k_dbt, k_histdbt;

    Defrec rec;
    Histrec hrec;
    int account, branch, teller;

    //
    // XXX We could move a lot of this into the driver to make this
    // faster.
    //
    account = random_id(ACCOUNT, anum, bnum, tnum);
    branch = random_id(BRANCH, anum, bnum, tnum);
    teller = random_id(TELLER, anum, bnum, tnum);

    k_dbt.setSize(sizeof(int));

    d_dbt.setReturnInUserBuffer(1);
    d_dbt.setData(&rec);
    d_dbt.setUserBufferSize(sizeof(rec));

    hrec.aid = account;
    hrec.bid = branch;
    hrec.tid = teller;
    hrec.amount = 10;
    // Request 0 bytes since we're just positioning.
    d_histdbt.setPartial(1);

    // START TIMING
    DBTransaction t(txmgr);
    if (!t.isValid())
        goto err;

    if (adb.getIterator(&acurs, t) != 0 ||
        bdb.getIterator(&bcurs, t) != 0 ||
        tdb.getIterator(&tcurs, t) != 0 ||
        hdb.getIterator(&hcurs, t) != 0)
        goto err;

    // Account record
    k_dbt.setData(&account);
    if (acurs.get(&k_dbt, &d_dbt) != 0)
        goto err;
    rec.balance += 10;
    if (acurs.putCurrent(d_dbt) != 0)
        goto err;

    // Branch record
    k_dbt.setData(&branch);
    if (bcurs.get(&k_dbt, &d_dbt) != 0)
        goto err;
    rec.balance += 10;
    if (bcurs.putCurrent(k_dbt) != 0)
        goto err;

    // Teller record
    k_dbt.setData(&teller);
    if (tcurs.get(&k_dbt, &d_dbt) != 0)
        goto err;
    rec.balance += 10;
    if (tcurs.putCurrent(k_dbt) != 0)
        goto err;

    // History record
    hcurs.setToEnd();
    if (hcurs.get(&k_histdbt, &d_histdbt) != 0)
        goto err;
    d_histdbt.setData(&hrec);
    d_histdbt.setUserBufferSize(sizeof(hrec));
    if (hcurs.putAfter(k_histdbt, d_histdbt) != 0)
        goto err;

    if (acurs.close() != 0 || bcurs.close() != 0 ||
        tcurs.close() != 0 || hcurs.close() != 0)
        goto err;

    if (t.commit() != 0)
        goto err;

    // END TIMING
    return (0);

err:
    if (acurs.isValid())
        (void)acurs.close();
    if (bcurs.isValid())
        (void)bcurs.close();
    if (tcurs.isValid())
        (void)tcurs.close();
    if (hcurs.isValid())
        (void)hcurs.close();
    if (t.isValid())
        (void)t.abort();

    if (verbose)
        cout << "Transaction A=" << (long)account
             << " B=" << (long)branch
             << " T=" << (long)teller << " failed\n";
    return (-1);
}

void errExit(int err, const char *s)
{
    cerr << progname << ": ";
    if (s != NULL) {
        cerr << s << ": ";
    }
    cerr << strerror(err) << "\n";
    exit(1);
}
