/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1997
 *	Sleepycat Software.  All rights reserved.
 *
 *	@(#)cxx_int.h	10.3 (Sleepycat) 7/20/97
 */

#ifndef _CXX_INT_H_
#define _CXX_INT_H_

// private data structures known to the implementation only

#include <assert.h>             // used by defines below

//
// Using FooImp classes will allow the implementation to change in the
// future without any modification to user code or even to header files
// that the user includes. FooImp * is just like void * except that it
// provides a little extra protection, since you cannot randomly assign
// any old pointer to a FooImp* as you can with void *.  Currently, a
// pointer to such an opaque class is always just a pointer to the
// appropriate underlying implementation struct.  These are converted
// back and forth using the various overloaded wrap()/unwrap() methods.
// This is essentially a use of the "Bridge" Design Pattern.
//
// WRAPPED_CLASS implements the appropriate wrap() and unwrap() methods
// for a wrapper class that has an underlying pointer representation.
//
// WRAPPED_CLASS_USING_TYPE implements the appropriate wrap() and
// unwrap() methods for a wrapper class that has an underlying basic type
// representation (such as int).
//
#define WRAPPED_CLASS_WITH_NAMES(_WRAPPER_CLASS, _IMP_CLASS, _WRAPPED_TYPE, \
                                 _WRAP_NAME, _UNWRAP_NAME, _UNWRAP_CNAME) \
        class _IMP_CLASS {}; \
        inline _WRAPPED_TYPE _UNWRAP_NAME(_WRAPPER_CLASS *val) \
            { return (_WRAPPED_TYPE)(val->imp()); } \
        inline const _WRAPPED_TYPE _UNWRAP_CNAME(const _WRAPPER_CLASS *val) \
            { return (const _WRAPPED_TYPE)(val->imp()); } \
        inline _IMP_CLASS *_WRAP_NAME(_WRAPPED_TYPE val) \
            { return (_IMP_CLASS*)val; }

#define WRAPPED_CLASS(_WRAPPER_CLASS, _IMP_CLASS, _WRAPPED_TYPE) \
        WRAPPED_CLASS_WITH_NAMES(_WRAPPER_CLASS, _IMP_CLASS, _WRAPPED_TYPE, \
                                 wrap, unwrap, unwrapConst)

#define WRAPPED_CLASS_USING_TYPE(_WRAPPER_CLASS, _WRAPPED_TYPE) \
        inline _WRAPPED_TYPE unwrap(_WRAPPER_CLASS *val) \
            { return (_WRAPPED_TYPE)(val->imp()); } \
        inline const _WRAPPED_TYPE unwrapConst(const _WRAPPER_CLASS *val) \
            { return (const _WRAPPED_TYPE)(val->imp()); } \

WRAPPED_CLASS(DBApp, DBAppImp, DB_ENV*)
WRAPPED_CLASS_WITH_NAMES(DBAppOptions, DBAppOptionsImp, DB_ENV*,
                         wrapOptions, unwrapOptions, unwrapOptionsConst)
WRAPPED_CLASS(DBIterator, DBIteratorImp, DBC*)
WRAPPED_CLASS(DBLockRequest, DBLockRequestImp, DB_LOCKREQ*)
WRAPPED_CLASS(DBLockTable, DBLockTableImp, DB_LOCKTAB*)
WRAPPED_CLASS(DBLog, DBLogImp, DB_LOG*)
WRAPPED_CLASS(DBMemPool, DBMemPoolImp, DB_MPOOL*)
WRAPPED_CLASS(DBMemPoolFile, DBMemPoolFileImp, DB_MPOOLFILE*)
WRAPPED_CLASS(DBOpenParams, DBOpenParamsImp, DB_INFO*)
WRAPPED_CLASS(DBSequenceNumber, DBSequenceNumberImp, DB_LSN*)
WRAPPED_CLASS(DBTable, DBTableImp, DB*)
WRAPPED_CLASS(DBThunk, DBThunkImp, DBT*)
WRAPPED_CLASS(DBTransaction, DBTransactionImp, DB_TXN*)
WRAPPED_CLASS(DBTransactionManager, DBTransactionManagerImp, DB_TXNMGR*)
WRAPPED_CLASS_USING_TYPE(DBLock, unsigned int)

// Macros that handle detected errors, in case we want to
// change the default behavior.  runtimeError() throws an
// exception by default.
//
// Since it's unusual to throw an exception in a destructor,
// we have a separate macro.  For now, we silently ignore such
// detected errors.
//
#define DB_ERROR(caller, ecode) \
    DBApp::runtimeError(caller, ecode)

#define DB_DESTRUCTOR_ERROR(caller, ecode) \
    DBApp::runtimeError(caller, ecode, 1)


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// These defines are for tedious flag or field set/get access methods.
//

// Define setName() and getName() methods that twiddle
// the _flags field.
//
#define DB_FLAG_METHODS(_class, _flags, _cxx_name, _flag_name) \
                                                               \
void _class::set##_cxx_name(int onOrOff)                       \
{                                                              \
    if (onOrOff)                                               \
        _flags |= _flag_name;                                  \
    else                                                       \
        _flags &= ~(_flag_name);                               \
}                                                              \
                                                               \
int _class::get##_cxx_name() const                             \
{                                                              \
    return (_flags & _flag_name) ? 1 : 0;                      \
}

// Define setName() and getName() methods that use the
// the _c_name field within _c_struct.
//
#define DB_ACCESS_METHODS(_class, _type, _cxx_name, _c_struct, _c_name)   \
void _class::set##_cxx_name(_type value)                       \
{                                                              \
    _c_struct *p = unwrap(this);                               \
    assert(p != 0);                                            \
    p->_c_name = value;                                        \
}                                                              \
                                                               \
_type _class::get##_cxx_name() const                           \
{                                                              \
    const _c_struct *p = unwrapConst(this);                    \
    assert(p != 0);                                            \
    return p->_c_name;                                         \
}

#define DB_ACCESS_METHODS_WITH_FLAG(_class, _type, _cxx_name,  \
                                    _c_struct, _c_name, flag)  \
void _class::set##_cxx_name(_type value)                       \
{                                                              \
    _c_struct *p = unwrap(this);                               \
    assert(p != 0);                                            \
    p->_c_name = value;                                        \
    p->flags |= (flag);                                        \
}                                                              \
                                                               \
_type _class::get##_cxx_name() const                           \
{                                                              \
    const _c_struct *p = unwrapConst(this);                    \
    assert(p != 0);                                            \
    return p->_c_name;                                         \
}                                                              \
                                                               \
int _class::has##_cxx_name() const                             \
{                                                              \
    const _c_struct *p = unwrapConst(this);                    \
    assert(p != 0);                                            \
    return ((p->flags & (flag)) != 0);                         \
}
#endif /* !_CXX_INT_H_ */
